@@ -10,10 +10,14 @@ import os
 import glob
 import re
 import sets
+import sys
+import traceback
 
 #
 # General utilities
 #
+
+
 def get_file(option):
     """Get the file corresponding to the given option
     """
@@ -150,103 +154,176 @@ def command_name():
                                                       "Ide",
                                                       "adacontrol") or "adactl"
 
+#
+# Management of fixes
+#
+class Fix:
+    def apply(self):
+        start_loc = self.fix_start.location()
+        end_loc = self.fix_end.location()
+        if self.key == "Delete":
+            start_loc.buffer().delete(start_loc, end_loc)
+        elif self.key in ["Replace", "Insert"]:
+            if self.key == "Replace":
+                start_loc.buffer().delete(start_loc, end_loc)
+
+            # There is at least one line of data, otherwise use Delete
+            start_loc.buffer().insert(start_loc, self.fix_data[0])
+            for line in self.fix_data[1:]:
+                end_loc.buffer().insert(end_loc, "\n" + line)
+        elif self.key == "Refactor":
+            GPS.EditorView(start_loc.buffer()).goto(start_loc)
+            GPS.Action("rename entity").execute_if_possible()
+
+
+def fixer(self):
+    """ Method of a message called when clicking the wrench
+    """
+    for fix in self.fix_list:
+        fix.apply()
+    self.set_subprogram(no_action, '', '')
+
+def no_action(self):
+    """ placeholder when the fix has been performed
+    """
+    pass
+
+def apply_fixes ():
+    """ Automatically perform all possible fixes
+    """
+    global adactl_cats
+
+    for cat in adactl_cats:
+        for mess in GPS.Message.list (category=cat):
+            mess.execute_action()
 
 def parse(output):
     """Sort and parse the result of running Adacontrol
     """
-    global adactl_cats, long_mess_str, long_mess_pat, short_mess_pat, \
-        rule_statistics_pat, message_statistics_pat
-
-    def sloc_cmp(l, r):
-        """Compare two source references (file:line:col)
-           If the references are the same, keep the original order
-           (for the case of error messages that span several lines)
-        """
-        ls = l.split(':')
-        rs = r.split(':')
-        if len(ls[0]) == 1:
-            # Presumably, we have a "C:"...
-            Offset = 1
-        else:
-            Offset = 0
-
-        if len(ls) < 5 + Offset and len(rs) < 3 + Offset:
-            # No reference =>Keep order
-            return -1
-        if len(ls) < 5 + Offset:
-            # Reference only in r: put references ahead
-            return 1
-        if len(rs) < 5 + Offset:
-            # Reference only in l: put references ahead
-            return -1
-
-        if ls[0 + Offset] != rs[0 + Offset]:
-            return cmp(ls[0 + Offset], rs[0 + Offset])
-        if ls[1 + Offset] != rs[1 + Offset]:
-            return cmp(int(ls[1 + Offset]), int(rs[1 + Offset]))
-        return cmp(int(ls[2 + Offset]), int(rs[2 + Offset]))
+    global adactl_cats, long_mess_pat, short_mess_pat, fix_mess_pat, \
+        rule_statistics_pat, message_statistics_pat, errors_style, warnings_style
 
     list = output.splitlines()
-    list.sort(sloc_cmp)
     pos = 0
     category = "Adacontrol"
     sep = ';'
     for Mess in list:
         pos += 1
-        if long_mess_pat.match(Mess):
-            # Regular AdaCtl message
+
+        if Mess == "":
+            continue
+
+        # Regular AdaCtl message
+        match = long_mess_pat.match(Mess)
+        if match:
             if GPS.Preference("separate-rules").get():
-                category = long_mess_pat.sub(r"\8", Mess)
+                category = match.group('rule')
             adactl_cats.add(category)
-            GPS.Locations.parse(Mess,
-                                category,
-                                long_mess_str,
-                                1, 2, 3, 4, 6, -1,
-                                "Builder results",
-                                "Style errors",
-                                "Builder warnings")
-        elif short_mess_pat.match(Mess):
-            # AdaCtl message without location, or just a column location
-            # (syntax error in interactive command)
+            mess = GPS.Message(category,
+                               file=GPS.File(match.group('file')),
+                               line=int(match.group('line')),
+                               column=int(match.group('col')),
+                               text=match.group('message'))
+            key = match.group('key')
+            if key == "Found":
+                mess.set_style(warnings_style)
+            else:
+                mess.set_style(errors_style)
+            prev_mess = mess
+            prev_mess.fix_list = []
+            continue
+
+        # Fix message
+        match = fix_mess_pat.match(Mess)
+        if match:
+            key = match.group('key')
+            buf = GPS.EditorBuffer.get(GPS.File(match.group('file')))
+            loc = GPS.EditorLocation(buf,
+                                     int(match.group('startline')),
+                                     int(match.group('startcol')))
+            fix = Fix()
+            fix.key = key
+            fix.fix_start = loc.create_mark(left_gravity=False)
+            fix.fix_data = []
+            if key == "Delete":
+                col = int(match.group('endcol'))
+                # col=0 means that the preceding end of line mark is to be deleted
+                if col == 0:
+                    loc = GPS.EditorLocation(buf,
+                                             int(match.group('endline'))-1,
+                                             1)
+                    loc = loc.end_of_line()
+                else:
+                    loc = GPS.EditorLocation(buf,
+                                             int(match.group('endline')),
+                                             col)
+                fix.fix_end = loc.create_mark(left_gravity=True)
+                prev_mess.set_subprogram(fixer, 'gps-codefix', "Delete it")
+            elif key == "Replace":
+                loc = GPS.EditorLocation(buf,
+                                         int(match.group('endline')),
+                                         int(match.group('endcol')))
+                fix.fix_end = loc.create_mark(left_gravity=True)
+                prev_mess.set_subprogram(fixer, 'gps-codefix', "Replace it")
+            elif key == "Refactor":
+                loc = GPS.EditorLocation(buf,
+                                         int(match.group('endline')),
+                                         int(match.group('endcol')))
+                fix.fix_end = loc.create_mark(left_gravity=True)
+                prev_mess.set_subprogram(fixer, 'gps-codefix', "Refactor it")
+            elif key == "Insert":
+                fix.fix_end = fix.fix_start
+                prev_mess.set_subprogram(fixer, 'gps-codefix', "Insert it")
+            prev_mess.fix_list.append(fix)
+            continue
+
+        # Fix data
+        if Mess[0] == '!':
+            prev_mess.fix_list[-1].fix_data.append(Mess[1:])
+            continue
+
+        # AdaCtl message without location, or just a column location
+        # (syntax error in interactive command)
+        match = short_mess_pat.match(Mess)
+        if match:
             if GPS.Preference("separate-rules").get():
-                category = short_mess_pat.sub(r"\3", Mess)
+                category = match.group('category')
             adactl_cats.add(category)
-            message = short_mess_pat.sub(
-                r"\2", Mess) + short_mess_pat.sub(r"\3", Mess)
+            message = match.group('message')
             try:
-                GPS.Locations.add(category, GPS.File("none"), 1, 1, message)
+                mess = GPS.Message(category, GPS.File("none"), 1, 1, message)
             except:
                 # Always an exception, since file "none" does not exist
                 # (presumably)
                 pass
-        elif rule_statistics_pat.match(Mess):
-            # Rule statistics summary
+            continue
+
+        # Rule statistics summary
+        match = rule_statistics_pat.match(Mess)
+        if match:
             adactl_cats.add("Statistics")
             try:
-                GPS.Locations.add("Statistics",
-                                  GPS.File(
-                                      rule_statistics_pat.sub(
-                                          r"\1", Mess)), 1, 1,
-                                  rule_statistics_pat.sub(r"\2", Mess))
+                mess = GPS.Message("Statistics", GPS.File(match.group('rule')), 1, 1, match.group('counts'))
             except:
                 # Always an exception, since file does not exist
                 pass
-        elif message_statistics_pat.match(Mess):
-            # Message statistics summary
+            continue
+
+        # Message statistics summary
+        match = message_statistics_pat.match(Mess)
+        if match:
             adactl_cats.add("Statistics")
             try:
-                GPS.Locations.add(
-                    "Statistics",
-                    GPS.File(message_statistics_pat.sub(r"\1", Mess)), 1, 1,
-                    message_statistics_pat.sub(r"\2", Mess))
+                mess = GPS.Message("Statistics", GPS.File(match.group('title')), 1, 1, match.group('counts'))
             except:
                 # Always an exception, since file does not exist
                 pass
-        else:
-            # Assume it is a compilation message
-            if Mess not in list[max(0, pos - 5):pos - 1]:
-                GPS.Locations.parse(Mess, "Compilation")
-                adactl_cats.add("Compilation")
+            continue
+
+        # Assume it is a compilation message
+        if Mess not in list[max(0, pos - 5):pos - 1]:
+            adactl_cats.add("Compilation")
+            GPS.Locations.parse(Mess, "Compilation")
 
 
 def load_result(file=""):
@@ -281,6 +358,8 @@ def load_result(file=""):
     except:
         GPS.MDI.dialog(
             "File " + previous_locfile + " is not a valid result file")
+        print (sys.exc_info())
+        traceback.print_tb (sys.exc_info()[2])
         return
     GPS.MDI.get("Locations").raise_window()
 
@@ -319,6 +398,9 @@ def options(rules, files):
         except:
             GPS.MDI.dialog("no active window")
             raise ValueError
+        if win == None :
+            GPS.MDI.dialog("no active window")
+            raise ValueError
         if win.language() != "ada":
             GPS.MDI.dialog("active window is not Ada")
             raise ValueError
@@ -337,7 +419,7 @@ def options(rules, files):
             res = run("check", previous_command)
             if res != 0:
                 raise ValueError
-            result = result + " -l \"" + \
+            result = result + " -f -l \"" + \
                 re.sub('"', '~', previous_command) + "\""
         else:
             raise ValueError
@@ -358,11 +440,12 @@ def options(rules, files):
                 raise ValueError
             result = result + " -f " + win.name()
         else:
-            result = result + " -l \"" + re.sub('"', '~', files) + "\""
+            result = result + " -f -l \"" + re.sub('"', '~', files) + "\""
 
     # set other_params
     gps = True
     outfile = False
+    fixes = False
     output_name = ""
     ASIS_option = ""
 
@@ -371,6 +454,8 @@ def options(rules, files):
     next_is_ofile = False
     next_is_pfile = False
     next_is_ASIS = False
+    next_is_format = False
+    next_is_fixes = False
     for O in opt_list:
         if len(O) == 0:
             pass
@@ -382,6 +467,12 @@ def options(rules, files):
         elif next_is_ASIS:
             if rules != "check":
                 ASIS_option = O
+        elif next_is_format:
+            result = result + " -F " + O
+            next_is_format = False
+        elif next_is_fixes:
+            result = result + " -G " + O
+            next_is_fixes = False
         elif skip_next:
             skip_next = False
         elif O[0] == "@":
@@ -399,17 +490,29 @@ def options(rules, files):
             outfile = True
         elif O == "-f":
             skip_next = True
-        elif O == "-F" and gps:
-            skip_next = True
+        elif O == "-F":
+            next_is_format = True
         elif O[0:2] == "-F":
             # Bug in GPS: Combo does not add separator
-            if not gps:
+            if gps:
+                result = result + " -F gnat_short"
+            else:
                 result = result + ' ' + "-F " + O[2:]
+        elif O == "-G":
+            fixes = True
+            next_is_fixes = True
+        elif O[0:2] == "-G":
+            # Bug in GPS: Combo does not add separator
+            fixes = True
+            result = result + ' ' + "-G " + O[2:]
         elif O == "-o":
             next_is_ofile = True
         elif O == "-p":
             # compatibility: ignore -p option and following parameter
             next_is_pfile = True
+        elif O == "-P":
+            print "-P"
+            next_is_fixfile = True
         elif O[0:2] == "-S" and O[2:] != "":
             # Bug in GPS: Combo does not add separator
             result = result + ' ' + "-S " + O[2:]
@@ -443,6 +546,9 @@ def options(rules, files):
         # file is specified but not used
         output_name = ""
 
+    if gps and not fixes:
+        result = result + " -G check"
+
     if ASIS_option:
         result = result + " -- " + ASIS_option
 
@@ -452,8 +558,8 @@ def options(rules, files):
 def process_line(process, matching, rest):
     global result
     if GPS.Preference("display-run").get():
-        print matching
-    result = result + '\n' + matching
+        print matching,
+    result = result + matching
 
 
 def run(rules, files):
@@ -467,6 +573,7 @@ def run(rules, files):
     try:
         command_line = command_name() + ' ' + options(rules, files)
     except ValueError:
+        print "option error", rules, files
         return
 
     pre_clean()
@@ -475,6 +582,7 @@ def run(rules, files):
     proc = GPS.Process(command=command_line,
                        task_manager=True,
                        regexp="^.+$",
+                       single_line_regexp=True,
                        on_match=process_line,
                        progress_regexp="^\((\d+)/(\d+)\)",
                        progress_current=1,
@@ -579,14 +687,19 @@ def on_GPS_start(H):
        variables for example)
     """
     global adactl_cats, previous_command, previous_locfile, DelTree_Menu,\
-        long_mess_str, long_mess_pat, short_mess_pat, rule_statistics_pat,\
-        message_statistics_pat, gps_version
+        long_mess_pat, short_mess_pat, rule_statistics_pat, fix_mess_pat,\
+        message_statistics_pat, gps_version, errors_style, warnings_style
 
     # Global variables initialization
     adactl_cats = sets.Set()
     previous_command = ""
     previous_locfile = ""
 
+    styles_dict = {s.get_name(): s for s in GPS.Style.list()}
+    errors_style = styles_dict['Builder results']
+    warnings_style = styles_dict['Builder warnings']
+
+
     # Normalize version numbers to allow for correct ordering of (string) comparison
     # 6.1.1 => 06.1.1
     # 2016  => 16.0
@@ -597,17 +710,22 @@ def on_GPS_start(H):
     elif gps_version.find('.') == 1:
         gps_version = '0' + gps_version
 
-    # 1    2     3      456        7                           8
-    long_mess_str = r"^(.+):(\d+):(\d+): (((Found:)|(Error:|Parameter:|Syntax:))( .*?):.*)$"
-    long_mess_pat = re.compile(long_mess_str)
-    # 1       2                                  3
-    short_mess_pat = re.compile(
-        r"^(\d+: )?(Found:|Error:|Parameter:|Syntax:) (.*)$")
-    # 1     23                                     4
-    rule_statistics_pat = re.compile(
-        r"^(.*): ((Check: \d+, Search: \d+, Count: \d+)|(not triggered))$")
-    #                                      1                  2
-    message_statistics_pat = re.compile(r"^(Issued messages:) (.*)$")
+    # Message recognition patterns
+    long_mess_pat = re.compile(r'(?P<file>.+?):'
+                               r'(?P<line>\d+):'
+                               r'(?P<col>\d+): '
+                               r'(?P<message>(?P<key>Found|Error|Parameter|Syntax): (?P<rule>.*?):.*)$')
+
+    short_mess_pat = re.compile(r"^(?P<category>.+):(?: )?(?P<message>(?P<key>Found|Error|Parameter|Syntax): (.*))$")
+
+    fix_mess_pat = re.compile(r'"?(?P<file>.+?)"?:'
+                              r'(?P<startline>\d+):'
+                              r'(?P<startcol>\d+): '
+                              r'(?P<key>Delete|Insert|Replace|Refactor)(:(?P<endline>\d+):(?P<endcol>\d+))?$')
+
+    rule_statistics_pat = re.compile(r"^(?P<rule>.*): (?P<counts>(Check: \d+, Search: \d+, Count: \d+)|(not triggered))$")
+
+    message_statistics_pat = re.compile(r"^(?P<title>(Issued messages|Units processed|Lines processed):) (?P<counts>.*)$")
 
     # Actions and submenus are now parsed from Python because otherwise
     # filters involving Python actions do not work anymore
@@ -666,6 +784,10 @@ def on_GPS_start(H):
      <shell lang="Python" show-command="false" output="">adactl.run("check", "")</shell>
    </action>
 
+   <action name="Apply_Fixes">
+     <shell lang="Python" show-command="false" output="">adactl.apply_fixes()</shell>
+   </action>
+
    <action name="Get_Units" show-command="false" output="none">
      <filter shell_cmd="adactl.units_file_defined()" shell_lang="python" />
      <shell lang="python">adactl.open_file(adactl.get_units_file())</shell>
@@ -739,6 +861,12 @@ def on_GPS_start(H):
 
      <menu />
 
+     <menu action="Apply_Fixes">
+        <title>Apply all fixes</title>
+     </menu>
+
+     <menu />
+
      <menu action="Check_AdaCtl">
         <title>Check Rules File Syntax</title>
      </menu>
@@ -786,7 +914,8 @@ def on_GPS_start(H):
          <pixmap>{base_dir}share/gps/plug-ins/adactl-ask.gif</pixmap>
       </button>
       """ .format(base_dir=GPS.get_system_dir()))
-    elif gps_version <= "06.2.1" :
+
+    elif gps_version <= "06.2.1":
         # GPS 06.1.2 up to 06.2.1: Buttons use stock
         GPS.parse_xml('''
       <stock>
